@@ -4,9 +4,10 @@ import logging
 import numbers
 import threading
 import time
+import typing
 import uuid
 from inspect import signature
-from typing import Optional, Any, Callable, Mapping, Union, Iterable
+from typing import Optional, Any, Callable, Mapping, Union, Iterable, Type
 
 from pika import (BlockingConnection, BasicProperties, ConnectionParameters,
                   PlainCredentials)
@@ -17,12 +18,11 @@ from pika.exchange_type import ExchangeType
 from pika.frame import Method
 from pika.spec import Basic
 
-__all__ = ('RabbitClient',)
+__all__ = ('RabbitClient', 'BasicProperties')
 log = logging.getLogger('snacks')
 
 
 # TODO django_rest_framework serializer support?
-# TODO default request/response serializers?
 class RabbitClient:
     """A class to interface with RabbitMQ."""
 
@@ -30,20 +30,23 @@ class RabbitClient:
             self,
             host: Optional[str] = None,
             port: Optional[Union[str, int]] = None,
-            default_exchange: Optional[str] = None,
             virtual_host: Optional[str] = None,
             username: Optional[str] = None,
             password: Optional[str] = None,
-            retry_timeout: Optional[float] = None
+            default_exchange: Optional[str] = None,
+            default_content_type: Optional[str] = None,
+            retry_timeout: float = 10.0
     ) -> None:
         self.host = host or '127.0.0.1'
         self.port = port or 5762
         self.default_exchange = default_exchange or ''
+        self.default_content_type = default_content_type or ''
         self.virtual_host = virtual_host or '/'
         log.info('RabbitClient configured.')
         log.debug('host=%s', self.host)
         log.debug('port=%s', self.port)
         log.debug('default_exchange=%s', self.default_exchange)
+        log.debug('default_content_type=%s', self.default_content_type)
         log.debug('virtual_host=%s', self.virtual_host)
         if username and password:
             self.credentials = PlainCredentials(username, password)
@@ -57,7 +60,7 @@ class RabbitClient:
         )
         self.connection: BlockingConnection = BlockingConnection(self.params)
         self.channel: BlockingChannel = self.connection.channel()
-        self.retry_timeout: float = retry_timeout or 10.0
+        self.retry_timeout: float = retry_timeout
 
     @staticmethod
     def from_dict(dictionary: dict[str, Any]) -> 'RabbitClient':
@@ -216,10 +219,17 @@ class RabbitClient:
         """
         self._reset_if_lost()
         log.debug('Publishing key=%s msg=%s', routing_key, body)
+        try:
+            properties = kwargs.pop('properties')
+        except KeyError:
+            properties = BasicProperties(
+                content_type=self.default_content_type
+            )
         self.channel.basic_publish(
             exchange=exchange or self.default_exchange,
             routing_key=routing_key,
-            body=_def_serialize(body),
+            body=self._serialize(body),
+            properties=properties,
             **kwargs
         )
 
@@ -228,18 +238,16 @@ class RabbitClient:
             body: Any,
             routing_key: str,
             exchange: Optional[str] = None,
-            serialize: Optional[Callable] = None,
             time_limit: int = 60,
-            headers: Optional[dict[str, Any]] = None
+            **kwargs,
     ) -> Any:
         """Publish message to a rabbit queue with the given routing key.
 
         :param body: The message to publish.
         :param routing_key: The routing key.
         :param exchange: Exchange to publish to.
-        :param serialize: Callable to serialize response bodies.
         :param time_limit: Number of seconds to wait for a response.
-        :param headers: Headers to publish with.
+        :param kwargs: Keyword args to pass to pika publish.
         """
         self._reset_if_lost()
         log.debug('Publishing key=%s msg=%s', routing_key, body)
@@ -264,24 +272,28 @@ class RabbitClient:
             on_message_callback=_on_response,
             auto_ack=True
         )
+        try:
+            properties = kwargs.pop('properties')
+        except KeyError:
+            properties = BasicProperties(
+                reply_to=callback_queue,
+                correlation_id=corr_id,
+                content_type=self.default_content_type
+            )
         self.channel.basic_publish(
             exchange=exchange or self.default_exchange,
             routing_key=routing_key,
-            properties=BasicProperties(
-                reply_to=callback_queue,
-                correlation_id=corr_id,
-                headers=headers
-            ),
-            body=_def_serialize(body)
+            properties=properties,
+            body=self._serialize(body),
+            **kwargs
         )
         self.connection.process_data_events(time_limit=time_limit)
-        return serialize(response) if serialize else response
+        return response
 
     def listener(
             self,
             routing_keys: Union[list[str], str],
             exchange: Optional[str] = None,
-            serialize: Optional[Callable] = None,
             declare_arguments: Optional[Mapping[str, Any]] = None,
             bind_arguments: Optional[Mapping[str, Any]] = None
     ) -> Callable:
@@ -297,7 +309,6 @@ class RabbitClient:
 
         :param routing_keys: Key or keys to generate queues for.
         :param exchange: The queues to consume from.
-        :param serialize: Callable to serialize message bodies.
         :param declare_arguments: Arguments for queue_declare.
         :param bind_arguments: Arguments for queue_bind.
         :return: Function decorated to be a rabbit consumer.
@@ -332,7 +343,7 @@ class RabbitClient:
                 fun.__name__ = fun_name
             thread = threading.Thread(
                 target=self._consume,
-                args=(fun, [], serialize, queue_factory)
+                args=(fun, [], queue_factory)
             )
             thread.daemon = True
             thread.start()
@@ -346,11 +357,7 @@ class RabbitClient:
 
         return wrapper
 
-    def consumer(
-            self,
-            queues: Union[list[str], str],
-            serialize: Optional[Callable] = None
-    ) -> Callable:
+    def consumer(self, queues: Union[list[str], str]) -> Callable:
         """Decorate a callable to consume from one or more queues.
 
         The decorated function can have as parameters, any or all of,
@@ -360,15 +367,11 @@ class RabbitClient:
         consumer.
 
         :param queues: The queue or queues to consume from.
-        :param serialize: Callable to serialize message bodies.
         :return: Function decorated to be a rabbit consumer.
         """
 
         def wrapper(fun: Callable) -> Any:
-            thread = threading.Thread(
-                target=self._consume,
-                args=(fun, queues, serialize)
-            )
+            thread = threading.Thread(target=self._consume, args=(fun, queues))
             thread.daemon = True
             thread.start()
 
@@ -378,7 +381,6 @@ class RabbitClient:
             self,
             fun: Callable,
             queues: Union[Iterable[str], str],
-            serialize: Optional[Callable] = None,
             q_factory: Optional[Callable[[BlockingChannel], list[str]]] = None
     ) -> None:
         running = True
@@ -393,7 +395,7 @@ class RabbitClient:
                     channel = connection.channel()
 
                 def consume(q: str) -> None:
-                    _listen(channel, q, fun, serialize)
+                    _listen(channel, q, fun)
                     level = log.debug if q.startswith('amq.gen') else log.info
                     level('[%s] consuming queue [%s]', fun.__name__, q)
 
@@ -425,13 +427,21 @@ class RabbitClient:
         if self.channel.is_closed:
             self.channel = self.connection.channel()
 
+    def _serialize(self, body: Any) -> bytes:
+        if self.default_content_type == 'application/json':
+            if 'to_json' in dir(body):
+                return body.to_json().encode()
+            try:
+                return json.dumps(body).encode()
+            except TypeError:
+                return body
+        elif isinstance(body, (numbers.Number, Iterable)):
+            return str(body).encode()
+        else:
+            return body
 
-def _listen(
-        channel: BlockingChannel,
-        queue: str,
-        fun: Callable,
-        serialize: Optional[Callable] = None
-) -> None:
+
+def _listen(channel: BlockingChannel, queue: str, fun: Callable) -> None:
     sig = signature(fun)
 
     def callback(
@@ -441,36 +451,23 @@ def _listen(
             body: bytes
     ) -> None:
         log.debug('Queue [%s] received [%s]', queue, body)
+        log.debug('Properties [%s]', properties)
         kwargs: dict[str, Any] = {}
         for name, param in sig.parameters.items():
-            # FIXME Typing annotations break this.
-            #  If type is Optional, get rid of the Optional.
-            #  If type is Union try mapping to all, take what works.
-            #  If type is Iterable default to list.
-            #  If type is Mapping default to dict.
+            annotation = param.annotation
             if name == 'self':
                 kwargs[name] = None
-            elif param.annotation == BlockingChannel:
+            elif annotation == BlockingChannel:
                 kwargs[name] = ch
-            elif param.annotation == method:
+            elif annotation == method:
                 kwargs[name] = method
-            elif param.annotation == BasicProperties:
+            elif annotation == BasicProperties:
                 kwargs[name] = properties
-            elif param.annotation == str:
-                kwargs[name] = body.decode('utf-8')
-            elif param.annotation in {list, tuple, set, dict}:
-                kwargs[name] = ast.literal_eval(body.decode('UTF-8'))
-            elif param.annotation == bytes:
-                kwargs[name] = body
-            elif param.annotation in {int, float}:
-                kwargs[name] = param.annotation(body)
-            elif 'from_json' in dir(param.annotation):
-                kwargs[name] = param.annotation.from_json(body)
+            elif 'application/json' in [properties.content_type,
+                                        properties.content_encoding]:
+                kwargs[name] = _deserialize_json(body, annotation)
             else:
-                try:
-                    kwargs[name] = json.loads(body)
-                except json.decoder.JSONDecodeError:
-                    kwargs[name] = body
+                kwargs[name] = _deserialize(annotation, body)
         # noinspection PyBroadException
         try:
             resp = fun(**kwargs)
@@ -483,7 +480,7 @@ def _listen(
                     properties=BasicProperties(
                         correlation_id=properties.correlation_id
                     ),
-                    body=serialize(resp) if serialize else _def_serialize(resp)
+                    body=_serialize(resp)
                 )
                 ch.basic_ack(delivery_tag=method.delivery_tag)
         except Exception as e:
@@ -497,10 +494,55 @@ def _listen(
     )
 
 
-def _def_serialize(body: Any) -> bytes:
-    if 'to_json' in dir(body):
-        return body.to_json().encode()
-    elif isinstance(body, numbers.Number):
-        return str(body).encode()
-    else:
+def _deserialize(annotation: Type, body: bytes) -> Any:
+    # noinspection PyBroadException
+    try:
+        origin = typing.get_origin(annotation)
+        if annotation == str:
+            return body.decode('utf-8')
+        elif annotation == bytes:
+            return body
+        elif annotation in {int, float, numbers.Number}:
+            return annotation(body)
+        elif origin in {list, tuple, set, dict}:
+            return ast.literal_eval(body.decode('utf-8'))
+        elif origin == Union:
+            args = typing.get_args(annotation)
+            for arg in args:
+                b = _deserialize(arg, body)
+                if not isinstance(b, bytes):
+                    return b
+            return body
+        # TODO Arbitrary class serializing?
+        else:
+            return body
+    except Exception:
         return body
+
+
+def _deserialize_json(body: bytes, annotation: Type) -> Union[list[Any], Any]:
+    js_dict = json.loads(body)
+
+    def _deserialize_dict_list() -> list[Any]:
+        ret_val = []
+        processed = []
+        for arg in typing.get_args(annotation):
+            # noinspection PyBroadException
+            try:
+                for i, obj in enumerate(js_dict):
+                    ret_val.append(arg.from_dict(obj))
+                    processed.append(i)
+            except Exception:
+                [js_dict.pop(i) for i in processed]
+                processed = []
+        return ret_val if ret_val else js_dict
+
+    if isinstance(js_dict, dict) and 'from_json' in dir(annotation):
+        # noinspection PyUnresolvedReferences
+        return annotation.from_json(body)
+    elif isinstance(js_dict, list):
+        # FIXME Union[json_serializable_one, json_serializable_two]
+        if all('from_dict' in dir(a) for a in typing.get_args(annotation)):
+            return _deserialize_dict_list()
+    else:
+        return json.loads(body)
